<?php

namespace Drupal\Tests\oembed_lazyload_youtube\Kernel;

use Drupal\Core\Config\Config;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\KernelTests\KernelTestBase;
use Drupal\media\OEmbed\Provider;
use Drupal\media\OEmbed\Resource;
use Drupal\oembed_lazyload_youtube\Plugin\oembed_lazyload\ProviderEnhancer\YoutubeEnhancer;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Psr7\Request;
use GuzzleHttp\Psr7\Response;

/**
 * Test cases pertaining to the YouTube enhancer plugin.
 *
 * @group oembed_lazyload_youtube
 */
class YoutubeEnhancerTest extends KernelTestBase {

  /**
   * Data provider for YouTube video URLs.
   *
   * @see https://oembed.com/providers.json
   *
   * @return array
   *   An array of test URLs.
   */
  public function urlProvider() {

    return [
      ['https://www.youtube.com/watch?v=usN-pKfw6Q8'],
      ['https://www.youtube.com/watch?v=usN-pKfw6Q8&otherqueryparams'],
      ['https://www.youtube.com/watch?v=usN-pKfw6Q8#anchor'],
      ['https://www.youtube.com/watch?v=usN-pKfw6Q8&otherqueryparams#anchor'],

      ['https://youtu.be/usN-pKfw6Q8'],
      ['https://youtu.be/usN-pKfw6Q8?queryparams=test'],
      ['https://youtu.be/usN-pKfw6Q8#anchor'],
      ['https://youtu.be/usN-pKfw6Q8?queryparams=test#anchor'],

      ['https://www.youtube.com/v/usN-pKfw6Q8'],
      ['https://www.youtube.com/v/usN-pKfw6Q8?queryparams'],
      ['https://www.youtube.com/v/usN-pKfw6Q8#anchor'],
      ['https://www.youtube.com/v/usN-pKfw6Q8?queryparams#anchor'],

      ['https://www.youtube.com/playlist?list=usN-pKfw6Q8'],
      ['https://www.youtube.com/playlist?list=usN-pKfw6Q8?queryparams'],
      ['https://www.youtube.com/playlist?list=usN-pKfw6Q8#anchor'],
      ['https://www.youtube.com/playlist?list=usN-pKfw6Q8?queryparams#anchor'],

      ['https://youtube.com/playlist?list=usN-pKfw6Q8'],
      ['https://youtube.com/playlist?list=usN-pKfw6Q8?queryparams'],
      ['https://youtube.com/playlist?list=usN-pKfw6Q8#anchor'],
      ['https://youtube.com/playlist?list=usN-pKfw6Q8?queryparams#anchor'],

      ['https://www.youtube.com/shorts/usN-pKfw6Q8'],
      ['https://www.youtube.com/shorts/usN-pKfw6Q8?queryparams'],
      ['https://www.youtube.com/shorts/usN-pKfw6Q8#anchor'],
      ['https://www.youtube.com/shorts/usN-pKfw6Q8?queryparams#anchor'],
    ];
  }

  /**
   * Tests that thumbnails are built appropriately.
   *
   * @param string $url
   *   The URL to test.
   *
   * @dataProvider urlProvider
   *
   * @covers \Drupal\oembed_lazyload_youtube\Plugin\oembed_lazyload\ProviderEnhancer\YoutubeEnhancer::getPlaceholder
   */
  public function testGetPlaceholder($url) {

    $resource = $this->createMock(Resource::class);

    $provider = $this->createMock(Provider::class);

    $provider->method('getName')
      ->willReturn('YouTube');

    $resource->method('getProvider')
      ->willReturn($provider);

    $resource->method('getTitle')
      ->willReturn('Mocked video');

    $resource->method('getThumbnailUrl')
      ->willReturn('test thumbnail url');

    $expected = [
      '#theme' => 'oembed_lazyload_placeholder',
      '#url' => $url,
      '#title' => 'Mocked video',
      '#thumbnail' => 'test thumbnail url',
      '#settings' => ['test' => 'value', 'placeholder_asset_type' => 'remote_asset'],
      '#provider' => 'YouTube',
      '#third_party_settings' => ['embed_code' => 'usN-pKfw6Q8'],
    ];
    $enhancer = YoutubeEnhancer::create($this->container, [], 'youtube', []);

    $actual = $enhancer->getPlaceholder($url, $resource, ['test' => 'value', 'placeholder_asset_type' => 'remote_asset']);

    static::assertEquals($expected, $actual);
  }

  /**
   * Tests that the appropriate libraries are used.
   *
   * @covers \Drupal\oembed_lazyload_youtube\Plugin\oembed_lazyload\ProviderEnhancer\YoutubeEnhancer::getLibraries
   */
  public function testGetLibraries() {
    $expected = [
      'media/oembed.formatter',
      'oembed_lazyload/common',
      'oembed_lazyload_youtube/youtube',
    ];
    $enhancer = YoutubeEnhancer::create($this->container, [], 'youtube', []);
    static::assertEquals($expected, $enhancer->getLibraries());
  }

  /**
   * Tests that the appropriate alteration is performed to oembed responses.
   *
   * @covers \Drupal\oembed_lazyload_youtube\Plugin\oembed_lazyload\ProviderEnhancer\YoutubeEnhancer::alterOembedResponse
   */
  public function testAlterOembedResponse() {

    $config = $this->createMock(Config::class);
    $config->method('get')->willReturn('https://example.com');

    $config_factory = $this->createMock(ConfigFactoryInterface::class);
    $config_factory->method('get')->willReturn($config);

    $this->container->set('config.factory', $config_factory);

    $enhancer = YoutubeEnhancer::create($this->container, [], 'youtube', []);
    $markup = '<iframe width="200" height="113" src="https://www.youtube.com/embed/usN-pKfw6Q8?feature=oembed" allow="accelerometer; autoplay; clipboard-write; encrypted-media; gyroscope; picture-in-picture" allowfullscreen></iframe>';

    static::assertStringContainsString('src="https://www.youtube.com/embed/usN-pKfw6Q8?feature=oembed', $enhancer->alterOembedResponse($markup));

    // Ensure that setting the autoplay setting works.
    static::assertStringContainsString('src="https://www.youtube.com/embed/usN-pKfw6Q8?feature=oembed&autoplay=1', $enhancer->alterOembedResponse($markup, ['autoplay' => '1']));

    // Ensure that the modest branding setting works.
    static::assertStringContainsString('src="https://www.youtube.com/embed/usN-pKfw6Q8?feature=oembed&modestbranding=1', $enhancer->alterOembedResponse($markup, ['modestbranding' => '1']));

    // Ensure that the enablejsapi setting works.
    static::assertStringContainsString('src="https://www.youtube.com/embed/usN-pKfw6Q8?feature=oembed&enablejsapi=1', $enhancer->alterOembedResponse($markup, ['enablejsapi' => '1']));

    // Ensure that the origin setting does not set without enablejsapi.
    static::assertStringContainsString('src="https://www.youtube.com/embed/usN-pKfw6Q8?feature=oembed', $enhancer->alterOembedResponse($markup, ['origin' => '1']));

    // Ensure that the origin setting does set with enablejsapi.
    static::assertStringContainsString('src="https://www.youtube.com/embed/usN-pKfw6Q8?feature=oembed&enablejsapi=1&origin=https://example.com', $enhancer->alterOembedResponse($markup, [
      'enablejsapi' => '1',
      'origin' => '1',
    ]));

    // Ensure that the hideinfo setting works.
    static::assertStringContainsString('src="https://www.youtube.com/embed/usN-pKfw6Q8?feature=oembed&showinfo=0', $enhancer->alterOembedResponse($markup, ['hideinfo' => '1']));

    // Ensure that the rel setting works.
    static::assertStringContainsString('src="https://www.youtube.com/embed/usN-pKfw6Q8?feature=oembed&rel=0', $enhancer->alterOembedResponse($markup, ['rel' => '1']));

    // Throw everything all together.
    static::assertStringContainsString('src="https://www.youtube.com/embed/usN-pKfw6Q8?feature=oembed&autoplay=1&modestbranding=1&enablejsapi=1&origin=https://example.com&showinfo=0&rel=0', $enhancer->alterOembedResponse($markup, [
      'autoplay' => '1',
      'modestbranding' => '1',
      'enablejsapi' => '1',
      'origin' => '1',
      'hideinfo' => '1',
      'rel' => '1',
    ]));
  }

  /**
   * Data provider for thumbnail test URLs.
   *
   * @return array
   *   An array of thumbnail test URLs.
   */
  public function thumbnailUrlProvider() {
    return [
      ['https://www.youtube.com/watch?v=maxresdefault', 'maxresdefault', 1920, 1080],
      ['https://www.youtube.com/watch?v=hq720', 'hq720', 1280, 720],
      ['https://www.youtube.com/watch?v=sddefault', 'sddefault', 640, 480],
      ['https://www.youtube.com/watch?v=hqdefault', 'hqdefault', 480, 360],
      ['https://www.youtube.com/watch?v=mqdefault', 'mqdefault', 320, 180],
      ['https://www.youtube.com/watch?v=default', 'default', 120, 90],
    ];
  }

  /**
   * Tests that the appropriate alteration is performed to oembed metadata.
   *
   * @dataProvider thumbnailUrlProvider
   *
   * @covers \Drupal\oembed_lazyload_youtube\Plugin\oembed_lazyload\ProviderEnhancer\YoutubeEnhancer::alterOembedMetadata
   */
  public function testAlterOembedMetadata($url, $expected_format, $expected_width, $expected_height) {

    // Set up the configuration to allow alteration.
    $config = $this->createMock(Config::class);
    $config->method('get')->willReturn('prefer_max_resolution');
    $config_factory = $this->createMock(ConfigFactoryInterface::class);
    $config_factory->method('get')->willReturn($config);
    $this->container->set('config.factory', $config_factory);

    // Set up the HTTP client to only return "Success" for the expected format
    // while returning "Not Found" for the other requests.
    $http_client = $this->createMock(ClientInterface::class);
    $http_client
      ->method('request')
      ->willReturnCallback(function ($method, $url) use ($expected_format) {
        $response = $this->createMock(Response::class);
        $response->method('getStatusCode')
          ->willReturn($url === "https://i.ytimg.com/vi_webp/$expected_format/$expected_format.webp" ? 200 : 404);
        return $response;
      });
    $this->container->set('http_client', $http_client);

    // Ensure that the correct format is resolved given the static inputs.
    $actual = [
      'thumbnail_url' => "https://i.ytimg.com/vi/$expected_format/should-be-modified.jpg",
      'thumbnail_width' => 777,
      'thumbnail_height' => 777,
    ];
    $enhancer = YoutubeEnhancer::create($this->container, [], 'youtube', []);
    $enhancer->alterOembedMetadata($actual, $url);
    $expected = [
      'thumbnail_url' => "https://i.ytimg.com/vi_webp/$expected_format/$expected_format.webp",
      'thumbnail_width' => $expected_width,
      'thumbnail_height' => $expected_height,
    ];
    static::assertSame($expected, $actual);
  }

  /**
   * Tests that no alteration happens when the feature is disabled.
   *
   * @covers \Drupal\oembed_lazyload_youtube\Plugin\oembed_lazyload\ProviderEnhancer\YoutubeEnhancer::alterOembedMetadata
   */
  public function testAlterOembedMetadataDisabled() {

    // Without the "prefer_max_resolution" set, no HTTP requests should happen.
    $http_client = $this->createMock(ClientInterface::class);
    $http_client
      ->expects(static::never())
      ->method('request');
    $this->container->set('http_client', $http_client);

    $enhancer = YoutubeEnhancer::create($this->container, [], 'youtube', []);
    $actual = [
      'thumbnail_url' => 'https://i.ytimg.com/vi/-DYSucV1_9w/should-not-be-modified.jpg',
      'thumbnail_width' => 777,
      'thumbnail_height' => 777,
    ];
    $enhancer->alterOembedMetadata($actual, 'https://www.youtube.com/watch?v=default');
  }

  /**
   * Tests that no alteration happens when no remote assets are found.
   *
   * @covers \Drupal\oembed_lazyload_youtube\Plugin\oembed_lazyload\ProviderEnhancer\YoutubeEnhancer::alterOembedMetadata
   */
  public function testAlterOembedMetadataNoAssetsFound() {

    // Set up the configuration to allow alteration.
    $config = $this->createMock(Config::class);
    $config->method('get')->willReturn('prefer_max_resolution');
    $config_factory = $this->createMock(ConfigFactoryInterface::class);
    $config_factory->method('get')->willReturn($config);
    $this->container->set('config.factory', $config_factory);

    // Set up the HTTP client to always return "Not Found".
    $response = $this->createMock(Response::class);
    $response->method('getStatusCode')
      ->willReturn(404);

    $http_client = $this->createMock(ClientInterface::class);
    $http_client
      ->expects(static::exactly(6))
      ->method('request')
      ->willReturn($response);
    $this->container->set('http_client', $http_client);

    // Ensure that no alteration happens to the initial data.
    $expected = [
      'thumbnail_url' => 'https://i.ytimg.com/vi/-DYSucV1_9w/maxresdefault.jpg',
      'thumbnail_width' => 777,
      'thumbnail_height' => 777,
    ];
    $actual = $expected;
    $enhancer = YoutubeEnhancer::create($this->container, [], 'youtube', []);
    $enhancer->alterOembedMetadata($actual, 'https://www.youtube.com/watch?v=default');
    static::assertSame($expected, $actual);

  }

  /**
   * Tests that no alteration happens when the http client throws exceptions.
   *
   * @covers \Drupal\oembed_lazyload_youtube\Plugin\oembed_lazyload\ProviderEnhancer\YoutubeEnhancer::alterOembedMetadata
   */
  public function testAlterOembedMetadataClientException() {

    // Set up the configuration to allow alteration.
    $config = $this->createMock(Config::class);
    $config->method('get')->willReturn('prefer_max_resolution');
    $config_factory = $this->createMock(ConfigFactoryInterface::class);
    $config_factory->method('get')->willReturn($config);
    $this->container->set('config.factory', $config_factory);

    // Set up the HTTP client to always return "Not Found".
    $http_client = $this->createMock(ClientInterface::class);
    $http_client
      ->expects(static::exactly(6))
      ->method('request')
      ->willThrowException(new ClientException('Dummy exception', $this->createMock(Request::class), $this->createMock(Response::class)));
    $this->container->set('http_client', $http_client);

    // Ensure that no alteration happens to the initial data.
    $expected = [
      'thumbnail_url' => "https://i.ytimg.com/vi/-DYSucV1_9w/maxresdefault.jpg",
      'thumbnail_width' => 777,
      'thumbnail_height' => 777,
    ];
    $actual = $expected;
    $enhancer = YoutubeEnhancer::create($this->container, [], 'youtube', []);
    $enhancer->alterOembedMetadata($actual, 'https://www.youtube.com/watch?v=default');
    static::assertSame($expected, $actual);

  }

}
