<?php

namespace Drupal\oembed_lazyload\Plugin\Field\FieldFormatter;

use Drupal\Component\Utility\Html;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\FormatterBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\media\Entity\MediaType;
use Drupal\media\OEmbed\ResourceException;
use Drupal\media\Plugin\media\Source\OEmbedInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * A formatter plugin that lazy loads oembed video content.
 *
 * @FieldFormatter(
 *   id = "lazyload_oembed",
 *   label = @Translation("Lazy load oEmbed video"),
 *   field_types = {
 *     "link",
 *     "string",
 *     "string_long",
 *   },
 * )
 */
class LazyloadOEmbedFormatter extends FormatterBase {

  /**
   * The oEmbed resource fetcher.
   *
   * @var \Drupal\media\OEmbed\ResourceFetcherInterface
   */
  protected $resourceFetcher;

  /**
   * The oEmbed URL resolver service.
   *
   * @var \Drupal\media\OEmbed\UrlResolverInterface
   */
  protected $urlResolver;

  /**
   * The iFrame URL helper service.
   *
   * @var \Drupal\media\IFrameUrlHelper
   */
  protected $iFrameUrlHelper;

  /**
   * The oembed lazyload iFrame URL helper service.
   *
   * @var \Drupal\oembed_lazyload\IframeUrlHelper
   */
  protected $oembedLazyloadIframeUrlHelper;

  /**
   * The OEmbed lazyload manager service.
   *
   * @var \Drupal\oembed_lazyload\ProviderEnhancerManager
   */
  protected $oembedLazyloadPluginManager;

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The media settings config.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * The logger service.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);

    $instance->resourceFetcher = $container->get('media.oembed.resource_fetcher');
    $instance->urlResolver = $container->get('media.oembed.url_resolver');
    $instance->iFrameUrlHelper = $container->get('media.oembed.iframe_url_helper');
    $instance->oembedLazyloadIframeUrlHelper = $container->get('oembed_lazyload.iframe_url_helper');
    $instance->oembedLazyloadPluginManager = $container->get('oembed_lazyload');
    $instance->entityTypeManager = $container->get('entity_type.manager');
    $instance->moduleHandler = $container->get('module_handler');

    /** @var \Drupal\Core\Config\ConfigFactoryInterface $config_factory */
    $config_factory = $container->get('config.factory');
    $instance->config = $config_factory->get('media.settings');

    /** @var \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory */
    $logger_factory = $container->get('logger.factory');
    $instance->logger = $logger_factory->get('oembed_lazyload');

    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
      'dimensions' => 'absolute_maxima',
      'max_width' => 0,
      'max_height' => 0,
      'placeholder_loading' => 'lazy',
      'placeholder_asset_type' => 'remote_asset',
      'placeholder_local_asset_image_style_type' => \Drupal::moduleHandler()->moduleExists('responsive_image') ? 'responsive_image_style' : 'image_style',
      'placeholder_local_asset_responsive_image_style' => '',
      'placeholder_local_asset_image_style' => '',
      'strategy' => 'intersection_observer',
      'intersection_observer_margin' => '',
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $parents = [
      'fields',
      $this->fieldDefinition->getName(),
      'settings_edit_form',
      'settings',
    ];

    $form = parent::settingsForm($form, $form_state);
    $form['general'] = [
      '#type' => 'details',
      '#title' => $this->t('General settings'),
      '#open' => TRUE,
    ];

    $form['general']['dimensions_info'] = [
      '#type' => 'container',
      '#states' => [
        'visible' => [
          '[name="fields[field_media_oembed_video][settings_edit_form][settings][dimensions]"]' => ['value' => 'legacy'],
        ],
      ],
    ];

    $form['general']['dimensions_info']['messages'] = [
      '#theme' => 'status_messages',
      '#message_list' => [
        'warning' => [
          $this->t('The legacy option for dimensions is deprecated and will be removed version 3.0.0.'),
        ],
      ],
    ];

    $form['general']['dimensions'] = [
      '#type' => 'select',
      '#title' => $this->t('Dimensions are based on...'),
      '#options' => [
        'absolute_maxima' => $this->t('Maximum width and height (pixel based)'),
        'automatic' => $this->t('Placeholder image dimensions'),
        'legacy' => $this->t('Legacy (forced 16:9 aspect ratio)'),
      ],
      '#default_value' => $this->getSetting('dimensions'),
      '#parents' => array_merge($parents, ['dimensions']),
    ];

    $form['general']['absolute_maxima'] = [
      '#type' => 'container',
      '#title' => $this->t('Absolute maxima'),
      '#states' => [
        'visible' => [
          '[name="fields[field_media_oembed_video][settings_edit_form][settings][dimensions]"]' => ['value' => 'absolute_maxima'],
        ],
      ],
    ];

    $form['general']['absolute_maxima']['max_width'] = [
      '#type' => 'number',
      '#title' => $this->t('Maximum width'),
      '#description' => $this->t('Leave blank or enter 0 to use the resource width'),
      '#default_value' => $this->getSetting('max_width'),
      '#size' => 5,
      '#maxlength' => 5,
      '#field_suffix' => $this->t('pixels'),
      '#min' => 0,
      '#parents' => array_merge($parents, ['max_width']),
    ];

    $form['general']['absolute_maxima']['max_height'] = [
      '#type' => 'number',
      '#title' => $this->t('Maximum height'),
      '#description' => $this->t('Leave blank or enter 0 to use the resource height'),
      '#default_value' => $this->getSetting('max_height'),
      '#size' => 5,
      '#maxlength' => 5,
      '#field_suffix' => $this->t('pixels'),
      '#min' => 0,
      '#parents' => array_merge($parents, ['max_height']),
    ];

    $form['placeholder'] = [
      '#type' => 'details',
      '#title' => $this->t('Placeholder settings'),
    ];

    $form['placeholder']['info'] = [
      '#type' => 'container',
      '#states' => [
        'visible' => [
          '[name="fields[field_media_oembed_video][settings_edit_form][settings][placeholder_asset_type]"]' => ['value' => 'remote_asset'],
        ],
      ],
    ];

    $form['placeholder']['info']['messages'] = [
      '#theme' => 'status_messages',
      '#message_list' => [
        'warning' => [
          $this->t('The remote asset type option is deprecated and will be removed version 3.0.0.'),
        ],
      ],
    ];

    $form['placeholder']['placeholder_asset_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Asset type'),
      '#options' => [
        'remote_asset' => $this->t('Render a remote thumbnail asset'),
        'local_asset' => $this->t('Render a local thumbnail asset'),
      ],
      '#default_value' => $this->getSetting('placeholder_asset_type'),
      '#parents' => array_merge($parents, ['placeholder_asset_type']),
    ];

    $form['placeholder']['placeholder_local_asset'] = [
      '#type' => 'container',
      '#title' => $this->t('Local thumbnails'),
      '#states' => [
        'visible' => [
          '[name="fields[field_media_oembed_video][settings_edit_form][settings][placeholder_asset_type]"]' => ['value' => 'local_asset'],
        ],
      ],
    ];

    $form['placeholder']['placeholder_local_asset']['placeholder_local_asset_image_style_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Rendering mechanism'),
      '#options' => [
        'image_style' => $this->t('Image style'),
      ],
      '#default_value' => $this->getSetting('placeholder_local_asset_image_style_type'),
      '#parents' => array_merge($parents, ['placeholder_local_asset_image_style_type']),
    ];

    // The responsive image module may be uninstalled.
    if ($this->moduleHandler->moduleExists('responsive_image')) {
      $form['placeholder']['placeholder_local_asset']['placeholder_local_asset_image_style_type']['#options'] = [
        'responsive_image_style' => $this->t('Responsive image style'),
      ] + $form['placeholder']['placeholder_local_asset']['placeholder_local_asset_image_style_type']['#options'];

      $form['placeholder']['placeholder_local_asset']['placeholder_local_asset_responsive_image_style'] = [
        '#type' => 'select',
        '#title' => $this->t('Responsive image style'),
        '#options' => array_map(
          static fn (EntityInterface $entity) => $entity->label(),
          $this->entityTypeManager->getStorage('responsive_image_style')->loadMultiple()
        ),
        '#default_value' => $this->getSetting('placeholder_local_asset_responsive_image_style'),
        '#states' => [
          'required' => [
            '[name="fields[field_media_oembed_video][settings_edit_form][settings][placeholder_local_asset_image_style_type]"]' => ['value' => 'responsive_image_style'],
          ],
          'visible' => [
            '[name="fields[field_media_oembed_video][settings_edit_form][settings][placeholder_local_asset_image_style_type]"]' => ['value' => 'responsive_image_style'],
          ],
        ],
        '#parents' => array_merge($parents, ['placeholder_local_asset_responsive_image_style']),
      ];
    }

    $form['placeholder']['placeholder_local_asset']['local_asset']['placeholder_local_asset_image_style'] = [
      '#type' => 'select',
      '#title' => $this->t('Image style'),
      '#options' => array_map(
        static fn (EntityInterface $entity) => $entity->label(),
        $this->entityTypeManager->getStorage('image_style')->loadMultiple()
      ),
      '#default_value' => $this->getSetting('placeholder_local_asset_image_style'),
      '#states' => [
        'required' => [
          '[name="fields[field_media_oembed_video][settings_edit_form][settings][placeholder_local_asset_image_style_type]"]' => ['value' => 'image_style'],
        ],
        'visible' => [
          '[name="fields[field_media_oembed_video][settings_edit_form][settings][placeholder_local_asset_image_style_type]"]' => ['value' => 'image_style'],
        ],
      ],
      '#parents' => array_merge($parents, ['placeholder_local_asset_image_style']),
    ];

    $form['placeholder']['placeholder_loading'] = [
      '#type' => 'select',
      '#title' => $this->t('Asset loading mechanism'),
      '#options' => [
        'lazy' => $this->t('Lazy'),
        'eager' => $this->t('Eager'),
      ],
      '#default_value' => $this->getSetting('placeholder_loading'),
      '#parents' => array_merge($parents, ['placeholder_loading']),
    ];

    $form['lazyload_strategy'] = [
      '#type' => 'details',
      '#title' => $this->t('Iframe settings'),
    ];

    $form['lazyload_strategy']['onclick_info'] = [
      '#type' => 'container',
      '#states' => [
        'visible' => [
          'select[name="fields[field_media_oembed_video][settings_edit_form][settings][strategy]"]' => [
            'value' => 'onclick',
          ],
        ],
      ],
    ];

    $form['lazyload_strategy']['onclick_info']['messages'] = [
      '#theme' => 'status_messages',
      '#message_list' => [
        'status' => [
          $this->t('This strategy can be useful if users are required to give explicit consent to load third party content.'),
        ],
        'warning' => [
          $this->t('Mobile users may have to click play twice if their browser does not allow auto-play (notably, Chrome + Safari).'),
        ],
      ],
    ];

    $form['lazyload_strategy']['intersection_observer_info'] = [
      '#type' => 'container',
      '#states' => [
        'visible' => [
          'select[name="fields[field_media_oembed_video][settings_edit_form][settings][strategy]"]' => [
            'value' => 'intersection_observer',
          ],
        ],
      ],
    ];

    $form['lazyload_strategy']['intersection_observer_info']['messages'] = [
      '#theme' => 'status_messages',
      '#message_list' => [
        'status' => [
          $this->t('This strategy can be useful if users do not have to give explicit consent to load third party content.'),
        ],
      ],
    ];

    $form['lazyload_strategy']['strategy'] = [
      '#type' => 'select',
      '#title' => $this->t('Lazy loading strategy'),
      '#default_value' => $this->getSetting('strategy'),
      '#options' => [
        'onclick' => $this->t('Load third party assets when the user activates a play button'),
        'intersection_observer' => $this->t('Load third party assets when it enters the viewport'),
      ],
      '#parents' => array_merge($parents, ['strategy']),
    ];

    $form['lazyload_strategy']['intersection_observer_margin'] = [
      '#type' => 'textfield',
      '#title' => $this->t('How far away from the viewport should content begin to load?'),
      '#description' => $this->t('This value must be defined as a number of pixels or percentage, IE. 20px or 4%.'),
      '#default_value' => $this->getSetting('intersection_observer_margin'),
      '#element_validate' => [
        [
          get_class($this),
          'validateIntersectionObserverMargin',
        ],
      ],
      '#states' => [
        'visible' => [
          'select[name="fields[field_media_oembed_video][settings_edit_form][settings][strategy]"]' => [
            'value' => 'intersection_observer',
          ],
        ],
      ],
      '#parents' => array_merge($parents, ['intersection_observer_margin']),
    ];
    return $form;
  }

  /**
   * Validate callback for the intersection observer field.
   *
   * @param array $element
   *   The form being validated.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The state of the form.
   */
  public static function validateIntersectionObserverMargin(array $element, FormStateInterface $form_state) {
    $value = $element['#value'];
    if ($value !== '' && !preg_match('/^\d+(px|%)$/', $value)) {
      $form_state->setError($element, 'The margin must be a positive integer specified in px or %.');
    }
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = parent::settingsSummary();
    $dimensions = $this->getSetting('dimensions');
    if ($dimensions === 'absolute_maxima') {
      if ($this->getSetting('max_width') && $this->getSetting('max_height')) {
        $summary[] = $this->t('Maximum size: %max_width x %max_height pixels', [
          '%max_width' => $this->getSetting('max_width'),
          '%max_height' => $this->getSetting('max_height'),
        ]);
      }
      elseif ($this->getSetting('max_width')) {
        $summary[] = $this->t('Maximum width: %max_width pixels', [
          '%max_width' => $this->getSetting('max_width'),
        ]);
      }
      elseif ($this->getSetting('max_height')) {
        $summary[] = $this->t('Maximum height: %max_height pixels', [
          '%max_height' => $this->getSetting('max_height'),
        ]);
      }
      else {
        $summary[] = $this->t('Width and height are based on the underlying resource dimensions.');
      }
    }
    elseif ($dimensions === 'automatic') {
      $summary[] = $this->t('Width and height are based on placeholder image dimensions.');
    }
    else {
      $summary[] = $this->t('Rendering at a forced 16:9 aspect ratio.');
    }

    if ($this->getSetting('placeholder_loading')) {
      $summary[] = $this->t('Placeholder rendering with the %loading loading attribute.', [
        '%loading' => $this->getSetting('placeholder_loading'),
      ]);
    }

    if ($this->getSetting('placeholder_asset_type') === 'remote_asset') {
      $summary[] = $this->t('Placeholder asset will be loaded from a third party server.');
    }
    else {
      $summary[] = $this->t('Placeholder asset will be loaded from a local resource.');
      if ($this->getSetting('placeholder_local_asset_image_style_type') === 'responsive_image_style') {
        $summary[] = $this->t('Placeholder asset will be rendered with the %responsive_image_style responsive image style.', [
          '%responsive_image_style' => $this->getSetting('placeholder_local_asset_responsive_image_style'),
        ]);
      }
      else {
        $summary[] = $this->t('Placeholder asset will be rendered with the %image_style  image style.', [
          '%image_style' => $this->getSetting('placeholder_local_asset_image_style'),
        ]);
      }
    }

    if ($this->getSetting('strategy') === 'onclick') {
      $summary[] = $this->t('Third party assets load when a user clicks a play button');
    }
    elseif ($margin = $this->getSetting('intersection_observer_margin')) {
      $summary[] = $this->t('Third party assets will begin to load @unit before entering the viewport.', [
        '@unit' => $margin,
      ]);
    }
    else {
      $summary[] = $this->t('Third party assets load as soon as an asset enters the viewport');
    }
    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public static function isApplicable(FieldDefinitionInterface $field_definition) {
    $applicable = FALSE;
    if (parent::isApplicable($field_definition) && $field_definition->getTargetEntityTypeId() === 'media') {
      $media_type = $field_definition->getTargetBundle();

      if ($media_type) {
        $media_type = MediaType::load($media_type);
        $applicable = $media_type && $media_type->getSource() instanceof OEmbedInterface;
      }
    }
    return $applicable;
  }

  /**
   * {@inheritdoc}
   *
   * This is based on the internal OEmbedFormatter plugin.
   *
   * @see \Drupal\media\Plugin\Field\FieldFormatter\OEmbedFormatter::viewElements
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    $element = [];
    $settings = $this->getSettings();

    foreach ($items as $delta => $item) {

      /** @var \Drupal\Core\Field\FieldItemInterface $item */
      $main_property = $item->getFieldDefinition()->getFieldStorageDefinition()->getMainPropertyName();
      $value = $item->{$main_property};

      if (empty($value)) {
        continue;
      }

      try {
        $resource_url = $this->urlResolver->getResourceUrl($value, $settings['max_width'], $settings['max_height']);
        $resource = $this->resourceFetcher->fetchResource($resource_url);
      }
      catch (ResourceException $exception) {
        $this->logger->error('Could not retrieve the remote URL (@url).', ['@url' => $value]);
        continue;
      }
      $provider = $resource->getProvider();
      if ($provider === NULL) {
        $this->logger->error('Could not find the oembed provider.');
        continue;
      }
      /** @var \Drupal\oembed_lazyload\ProviderEnhancerInterface $enhancer */
      $enhancer = $this->oembedLazyloadPluginManager->getEnhancerForProvider($provider->getName());

      $query = [
        'url' => $value,
        'max_width' => $settings['max_width'],
        'max_height' => $settings['max_height'],
        'hash' => $this->iFrameUrlHelper->getHash($value, $settings['max_width'], $settings['max_height']),
        'oembed_lazyload' => 1,
        'provider' => $provider->getName(),
      ];

      $third_party_settings = $this->getThirdPartySettings($enhancer->getPluginDefinition()['provider']);
      if ($third_party_settings) {
        $query['options'] = $third_party_settings;
      }
      $query['oembed_lazyload_hash'] = $this->oembedLazyloadIframeUrlHelper->getHash($provider->getName(), $third_party_settings);

      $url = Url::fromRoute('media.oembed_iframe', [], [
        'query' => $query,
      ]);

      $domain = $this->config->get('iframe_domain');
      if ($domain) {
        $url->setOption('base_url', $domain);
      }

      $placeholder_loading = $this->getSetting('placeholder_loading');

      $element[$delta] = [
        '#theme' => 'oembed_lazyload',
        '#attributes' => [
          'class' => [
            'oembed-lazyload',
            'oembed-lazyload--' . Html::getClass($provider->getName()),
            'oembed-lazyload--' . Html::getClass($this->getSetting('dimensions')),
          ],
        ],
        '#attached' => [
          'library' => $enhancer->getLibraries(),
        ],
        '#thumbnail' => $resource->getThumbnailUrl(),
        '#strategy' => $this->getSetting('strategy'),
        '#provider' => $provider->getName(),
        '#placeholder' => $enhancer->getPlaceholder($value, $resource, $settings),
        '#placeholder_loading' => $placeholder_loading,
        '#iframe' => $enhancer->getIframe($url, $resource, $settings),
      ];

      // @todo Move this to ProviderEnhancerBase::getPlaceholder in 3.0.x.
      if ($this->getSetting('placeholder_asset_type') === 'local_asset') {
        $media = $item->getEntity();
        $file = $this->entityTypeManager->getStorage('file')->load($media->get('thumbnail')->target_id);
        if ($this->getSetting('placeholder_local_asset_image_style_type') === 'responsive_image_style') {
          $element[$delta]['#placeholder']['#native_thumbnail'] = [
            '#type' => 'responsive_image',
            '#responsive_image_style_id' => $this->getSetting('placeholder_local_asset_responsive_image_style'),
            '#uri' => $file->getFileUri(),
            '#attributes' => [
              'class' => 'oembed-lazyload__thumbnail',
              'alt' => $this->t('Watch @title on @provider.', [
                '@title' => $resource->getTitle(),
                '@provider' => $provider->getName(),
              ]),
            ],
          ];
          if ($placeholder_loading) {
            $element[$delta]['#placeholder']['#native_thumbnail']['#attributes']['loading'] = $placeholder_loading;
          }
        }
        else {
          $element[$delta]['#placeholder']['#native_thumbnail'] = [
            '#theme' => 'image_style',
            '#style_name' => $this->getSetting('placeholder_local_asset_image_style'),
            '#uri' => $file->getFileUri(),
            '#attributes' => [
              'class' => 'oembed-lazyload__thumbnail',
              'alt' => $this->t('Watch @title on @provider.', [
                '@title' => $resource->getTitle(),
                '@provider' => $provider->getName(),
              ]),
            ],
          ];
          if ($placeholder_loading) {
            $element[$delta]['#placeholder']['#native_thumbnail']['#attributes']['loading'] = $placeholder_loading;
          }
        }
      }

      if ($this->getSetting('strategy') === 'intersection_observer') {
        $element[$delta]['#attached']['library'][] = 'oembed_lazyload/intersection-observer';
        $element[$delta]['#attached']['drupalSettings']['intersectionObserverMargin'] = $this->getSetting('intersection_observer_margin');
      }
      else {
        $element[$delta]['#attached']['library'][] = 'oembed_lazyload/onclick';
      }

      CacheableMetadata::createFromObject($resource)
        ->addCacheTags($this->config->getCacheTags())
        ->applyTo($element[$delta]);
    }

    return $element;
  }

}
