<?php

namespace Drupal\Tests\oembed_lazyload\Functional;

use Drupal\Component\Utility\Crypt;
use Drupal\Core\Entity\Entity\EntityViewDisplay;
use Drupal\Core\Url;
use Drupal\image\Entity\ImageStyle;
use Drupal\media\Entity\Media;
use Drupal\media_test_oembed\Controller\ResourceController;
use Drupal\media_test_oembed\UrlResolver;
use Drupal\Tests\BrowserTestBase;
use Drupal\Tests\media\Traits\MediaTypeCreationTrait;
use Drupal\Tests\media\Traits\OEmbedTestTrait;

/**
 * Ensures that placeholder rendering logic works as expected.
 *
 * @group oembed_lazyload
 */
class LazyloadOEmbedFormatterPlaceholderTest extends BrowserTestBase {

  use OEmbedTestTrait;
  use MediaTypeCreationTrait;

  /**
   * The predictable path to a static fixture asset on disk.
   *
   * @var string
   */
  protected $file;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'media_test_oembed',
    'oembed_lazyload_test',
    'oembed_lazyload',
    'responsive_image',
    'field_ui',
    'toolbar',
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();
    $this->file = sprintf(
      "oembed_thumbnails/%s/%s.png",
      date('Y-m'),
      Crypt::hashBase64(Url::fromUri('internal:/core/misc/druplicon.png')->setAbsolute()->toString())
    );
    $this->lockHttpClientToFixtures();

    \Drupal::configFactory()
      ->getEditable('media.settings')
      ->set('standalone_url', TRUE)
      ->save(TRUE);

    $this->container->get('router.builder')->rebuild();
  }

  /**
   * Data provider with placeholder configurations.
   *
   * @return array[]
   *   All permutations of formatter placeholder configurations.
   */
  public function formatterSettingsProvider(): array {
    return [
      'Placeholder dimensions with remote assets' => [
        [
          'dimensions' => 'automatic',
          'placeholder_asset_type' => 'remote_asset',
        ],
        "<span class=\"oembed-lazyload__thumbnail\" style=\"background-image: url('{{BASE_URL}}/core/misc/druplicon.png')\"></span>",
      ],
      'Placeholder dimensions with local responsive image style' => [
        [
          'dimensions' => 'automatic',
          'placeholder_asset_type' => 'local_asset',
          'placeholder_local_asset_image_style_type' => 'responsive_image_style',
          'placeholder_local_asset_responsive_image_style' => 'test',
        ],
        '<img class="oembed-lazyload__thumbnail" loading="lazy" srcset="{{FILES_DIR}}/styles/test/public/{{FILE}}?itok={{ITOK}} 1x" width="88" height="100" src="data:image/gif;base64,R0lGODlhAQABAIABAP///wAAACH5BAEKAAEALAAAAAABAAEAAAICTAEAOw==" alt="Watch Drupal Rap Video - Schipulcon09 on Vimeo.">',
      ],
      'Placeholder dimensions with local image style' => [
        [
          'dimensions' => 'automatic',
          'placeholder_asset_type' => 'local_asset',
          'placeholder_local_asset_image_style_type' => 'image_style',
          'placeholder_local_asset_image_style' => 'test',
        ],
        '<img class="oembed-lazyload__thumbnail" alt="Watch Drupal Rap Video - Schipulcon09 on Vimeo." loading="lazy" src="{{FILES_DIR}}/styles/test/public/{{FILE}}?itok={{ITOK}}">',
      ],
      'Absolute maxima dimensions with remote assets' => [
        [
          'dimensions' => 'absolute_maxima',
          'placeholder_asset_type' => 'remote_asset',
        ],
        "<span class=\"oembed-lazyload__thumbnail\" style=\"background-image: url('{{BASE_URL}}/core/misc/druplicon.png')\"></span>",
      ],
      'Absolute maxima dimensions with local responsive image style' => [
        [
          'dimensions' => 'absolute_maxima',
          'placeholder_asset_type' => 'local_asset',
          'placeholder_local_asset_image_style_type' => 'responsive_image_style',
          'placeholder_local_asset_responsive_image_style' => 'test',
        ],
        '<img class="oembed-lazyload__thumbnail" loading="lazy" srcset="{{FILES_DIR}}/styles/test/public/{{FILE}}?itok={{ITOK}} 1x" width="88" height="100" src="data:image/gif;base64,R0lGODlhAQABAIABAP///wAAACH5BAEKAAEALAAAAAABAAEAAAICTAEAOw==" alt="Watch Drupal Rap Video - Schipulcon09 on Vimeo.">',
      ],
      'Absolute maxima dimensions with local image style' => [
        [
          'dimensions' => 'absolute_maxima',
          'placeholder_asset_type' => 'local_asset',
          'placeholder_local_asset_image_style_type' => 'image_style',
          'placeholder_local_asset_image_style' => 'test',
        ],
        '<img class="oembed-lazyload__thumbnail" alt="Watch Drupal Rap Video - Schipulcon09 on Vimeo." loading="lazy" src="{{FILES_DIR}}/styles/test/public/{{FILE}}?itok={{ITOK}}">',
      ],
      'Legacy dimensions with remote assets' => [
        [
          'dimensions' => 'legacy',
          'placeholder_asset_type' => 'remote_asset',
        ],
        "<span class=\"oembed-lazyload__thumbnail\" style=\"background-image: url('{{BASE_URL}}/core/misc/druplicon.png')\"></span>",
      ],
      'Legacy dimensions with local responsive image style' => [
        [
          'dimensions' => 'legacy',
          'placeholder_asset_type' => 'local_asset',
          'placeholder_local_asset_image_style_type' => 'responsive_image_style',
          'placeholder_local_asset_responsive_image_style' => 'test',
        ],
        '<img class="oembed-lazyload__thumbnail" loading="lazy" srcset="{{FILES_DIR}}/styles/test/public/{{FILE}}?itok={{ITOK}} 1x" width="88" height="100" src="data:image/gif;base64,R0lGODlhAQABAIABAP///wAAACH5BAEKAAEALAAAAAABAAEAAAICTAEAOw==" alt="Watch Drupal Rap Video - Schipulcon09 on Vimeo.">',
      ],
      'Legacy dimensions with local image style' => [
        [
          'dimensions' => 'legacy',
          'placeholder_asset_type' => 'local_asset',
          'placeholder_local_asset_image_style_type' => 'image_style',
          'placeholder_local_asset_image_style' => 'test',
        ],
        '<img class="oembed-lazyload__thumbnail" alt="Watch Drupal Rap Video - Schipulcon09 on Vimeo." loading="lazy" src="{{FILES_DIR}}/styles/test/public/{{FILE}}?itok={{ITOK}}">',
      ],
    ];
  }

  /**
   * Ensures that placeholder generation logic behaves as expected.
   *
   * @dataProvider formatterSettingsProvider
   */
  public function testPlaceholders(array $formatter_settings, string $expected_markup) {
    /** @var \Drupal\image\ImageStyleInterface $image_style */
    $image_style = ImageStyle::load('test');
    $expected_markup = str_replace(
      ['{{BASE_URL}}', '{{FILES_DIR}}', '{{FILE}}', '{{ITOK}}'],
      [
        $this->htmlOutputBaseUrl,
        \Drupal::request()->getBasePath() . '/' . $this->publicFilesDirectory,
        $this->file,
        $image_style->getPathToken("public://{$this->file}"),
      ],
      $expected_markup
    );
    $media_type = $this->createMediaType('oembed:video');

    $source = $media_type->getSource();
    $source_field = $source->getSourceFieldDefinition($media_type);

    EntityViewDisplay::create([
      'targetEntityType' => 'media',
      'bundle' => $media_type->id(),
      'mode' => 'full',
      'status' => TRUE,
    ])->removeComponent('thumbnail')
      ->setComponent($source_field->getName(), [
        'type' => 'lazyload_oembed',
        'settings' => $formatter_settings,
      ])
      ->save();

    $this->hijackProviderEndpoints();

    ResourceController::setResourceUrl('https://vimeo.com/7073899', $this->getFixturesDirectory() . '/video_vimeo.json');
    UrlResolver::setEndpointUrl('https://vimeo.com/7073899', 'video_vimeo.json');

    $entity = Media::create([
      'bundle' => $media_type->id(),
      'status' => 1,
      $source_field->getName() => 'https://vimeo.com/7073899',
    ]);
    $entity->save();

    $this->drupalGet($entity->toUrl());

    static::assertEquals(
      $expected_markup,
      $this->getSession()->getPage()->find('css', '.oembed-lazyload__thumbnail')->getOuterHtml()
    );
  }

}
