<?php

namespace Drupal\Tests\oembed_lazyload\Unit;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Cache\NullBackend;
use Drupal\Core\Config\Config;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\media\OEmbed\ProviderRepositoryInterface;
use Drupal\oembed_lazyload\Oembed\OembedLazyloadResourceFetcher;
use Drupal\oembed_lazyload\ProviderEnhancerInterface;
use Drupal\oembed_lazyload\ProviderEnhancerManagerInterface;
use Drupal\Tests\UnitTestCase;
use GuzzleHttp\Client;
use GuzzleHttp\Psr7\Response;

/**
 * Test cases pertaining to the service provider.
 *
 * @coversClass \Drupal\oembed_lazyload\Oembed\OembedLazyloadResourceFetcher
 *
 * @group oembed_lazyload
 */
class OEmbedLazyloadResourceFetcherTest extends UnitTestCase {

  /**
   * The mocked configuration object.
   *
   * @var \Drupal\Core\Config\Config|\PHPUnit\Framework\MockObject\MockObject
   */
  protected $mockConfig;

  /**
   * The mocked provider enhancer object.
   *
   * @var \Drupal\oembed_lazyload\ProviderEnhancerInterface|\PHPUnit\Framework\MockObject\MockObject
   */
  protected $mockProviderEnhancer;

  /**
   * The subject under test.
   *
   * @var \Drupal\oembed_lazyload\Oembed\OembedLazyloadResourceFetcher
   */
  protected $instance;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'oembed_lazyload',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();
    $this->mockConfig = $this->createMock(Config::class);
    $this->mockProviderEnhancer = $this->createMock(ProviderEnhancerInterface::class);

    $config_factory_mock = $this->createMock(ConfigFactoryInterface::class);
    $config_factory_mock
      ->method('get')
      ->willReturn($this->mockConfig);

    $provider_enhancer_manager_mock = $this->createMock(ProviderEnhancerManagerInterface::class);
    $provider_enhancer_manager_mock
      ->method('getEnhancerForProvider')
      ->willReturn($this->mockProviderEnhancer);

    $headers = [
      'Content-Type' => ['text/javascript'],
    ];
    $body = Json::encode([
      'version' => '1.0',
      'type' => 'video',
      'html' => 'test',
      'provider_name' => 'dummy',
    ]);

    $client = $this->createMock(Client::class);
    $client
      ->method('request')
      ->willReturn(new Response(200, $headers, $body));

    $this->instance = new OembedLazyloadResourceFetcher(
      $client,
      $this->createMock(ProviderRepositoryInterface::class),
      new NullBackend('default')
    );

    $this->instance->setConfigFactory($config_factory_mock);
    $this->instance->setProviderEnhancerManager($provider_enhancer_manager_mock);
  }

  /**
   * Test case for the resource fetcher when disabled.
   */
  public function testResourceFetcherDisabled() {
    $this->mockConfig
      ->method('get')
      ->with('enable_resource_fetcher_overrides')
      ->willReturn(FALSE);

    $this->mockProviderEnhancer
      ->expects(static::never())
      ->method('alterOembedMetadata');

    $this->instance
      ->fetchResource('https://example.com/oembed?url=resource');
  }

  /**
   * Test case for the resource fetcher when enabled.
   */
  public function testResourceFetcherEnabled() {
    $this->mockConfig
      ->method('get')
      ->with('enable_resource_fetcher_overrides')
      ->willReturn(TRUE);

    $this->mockProviderEnhancer
      ->expects(static::once())
      ->method('alterOembedMetadata');

    $this->instance
      ->fetchResource('https://example.com/oembed?url=resource');

  }

}
